/*
    mpg321 - a fully free clone of mpg123.
    Copyright (C) 2001 Joe Drew
    
    Originally based heavily upon:
    plaympeg - Sample MPEG player using the SMPEG library
    Copyright (C) 1999 Loki Entertainment Software
    
    Also uses some code from
    mad - MPEG audio decoder
    Copyright (C) 2000-2001 Robert Leslie
    
    Original playlist code contributed by Tobias Bengtsson <tobbe@tobbe.nu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include "mpg321.h"

enum mad_flow read_from_mmap(void *data, struct mad_stream *stream)
{
    buffer *playbuf = (buffer *)data;

    if(playbuf->done)
        return MAD_FLOW_STOP;

    if(playbuf->fd != -1)
    {
        fprintf(stderr, "read_from_mmap called when not expected!\n");
        exit(1);
    }
    
    if(!playbuf->buf)
    {
        fprintf(stderr, "read_from_mmap called with null buf!\n");
        exit(1);
    }

    mad_stream_buffer(stream, (unsigned char *)playbuf->buf, playbuf->length);
    
    playbuf->done = 1;

    return MAD_FLOW_CONTINUE;
}

/* assumes that playbuf->buf has been preallocated to BUF_SIZE */
enum mad_flow read_from_fd(void *data, struct mad_stream *stream)
{
    buffer *playbuf = data;
    
    if(playbuf->done)
        return MAD_FLOW_STOP;

    if(playbuf->fd == -1)
    {
        fprintf(stderr, "read_from_fd called when not expected!\n");
        exit(1);
    }
    
    if(!playbuf->buf)
    {
        fprintf(stderr, "read_from_fd called with null buf!\n");
    }
    
    if((playbuf->length = read(playbuf->fd, playbuf->buf, BUF_SIZE))==-1)
        playbuf->done = 1;
    
    mad_stream_buffer(stream, playbuf->buf, playbuf->length);
    
    return MAD_FLOW_CONTINUE;
}    
    
void open_ao_playdevice(struct mad_header const *header)
{
    int numChannels = MAD_NCHANNELS(header);
        if(options.opt & MPG321_USE_WAV)
        {
            int driver_id = ao_get_driver_id("wav");
            ao_option_t *ao_options = NULL;

            /* Don't have to check options.device here: we only define
               MPG321_USE_WAV when -w <wavfile> is defined, and <wavfile>
               is pointd to by options.device */
            ao_append_option(&ao_options, "file", options.device);

            if((playdevice=ao_open(driver_id, 16, header->samplerate, numChannels, ao_options))==NULL)
            {
                fprintf(stderr, "Error opening libao wav file driver.\n");
                exit(1);
            }
        }

        else if(options.opt & MPG321_USE_NULL)
        {
            int driver_id = ao_get_driver_id("null");
            if((playdevice = ao_open(driver_id, 0, 0, 0, NULL)) == NULL)
            {
                fprintf(stderr, "Error opening libao null driver.\n");
                exit(1);
            }
        }
        
        else if (options.opt & MPG321_USE_STDOUT)
        {
            ao_option_t * ao_options = NULL;
            int driver_id = ao_get_driver_id("raw");
        
            ao_append_option(&ao_options, "file", "-");
        
            if((playdevice=ao_open(driver_id, 16, header->samplerate, numChannels, ao_options))==NULL)
            {
                fprintf(stderr, "Error opening libao raw output driver.\n");
                exit(1);
            }
        }
        
        else if(options.opt & MPG321_USE_ESD)
        {
            ao_option_t *ao_options = NULL;
            int driver_id = ao_get_driver_id("esd");
           
            if(options.device)
                ao_append_option(&ao_options, "host", options.device);
        
            if((playdevice=ao_open(driver_id, 16, header->samplerate, numChannels, ao_options))==NULL)
            {
                fprintf(stderr, "Error opening libao esd driver.\n");
                exit(1);
            }
        }
        
        else if(options.opt & MPG321_USE_ALSA)
        {
            ao_option_t *ao_options = NULL;
            int driver_id = ao_get_driver_id("alsa");
            char *c;
           
            if (options.device)
            {
                if ((c = strchr(options.device, ':')) == NULL || strlen(c+1) < 1)
                {
                    fprintf(stderr, "Poorly formed ALSA card:device specification %s", options.device);
                    exit(1);
                }
                
                *(c++) = '\0'; /* change the : to a null to create two separate strings */
                
                ao_append_option(&ao_options, "card", options.device);
                ao_append_option(&ao_options, "dev", c);
            }
        
            if((playdevice=ao_open(driver_id, 16, header->samplerate, numChannels, ao_options))==NULL)
            {
                fprintf(stderr, "Error opening libao alsa driver.\n");
                exit(1);
            }
        }
        
        else
        {
            ao_option_t *ao_options = NULL;
            int driver_id = ao_get_driver_id(AUDIO_DEFAULT);
            
            if (strcmp(AUDIO_DEFAULT, "oss") == 0)
            {
                if (options.device)
                    ao_append_option(&ao_options, "dsp", options.device);
            }
            else if (strcmp(AUDIO_DEFAULT, "sun") == 0)
            {
                if (options.device)
                    ao_append_option(&ao_options, "dev", options.device);
            }

            if((playdevice=ao_open(driver_id, 16, header->samplerate, numChannels, ao_options))==NULL)
            {
                fprintf(stderr, "Error opening libao oss driver.\n");
                exit(1);
            }
        }
}

char * layerstring(enum mad_layer layer)
{
    switch(layer)
    {
        case MAD_LAYER_I:
            return "I";
        case MAD_LAYER_II:
            return "II";
        case MAD_LAYER_III:
            return "III";
        default:
            return "?";
    }
}
    
char * modestring(enum mad_mode mode)
{
    switch(mode)
    {
        case MAD_MODE_SINGLE_CHANNEL:
            return "single-channel";
        case MAD_MODE_DUAL_CHANNEL:
            return "dual-channel";
        case MAD_MODE_JOINT_STEREO:
            return "joint-stereo";
        case MAD_MODE_STEREO:
            return "stereo";
        default:
            return "?";
    }
}

char * modestringucase(enum mad_mode mode)
{
    switch(mode)
    {
        case MAD_MODE_SINGLE_CHANNEL:
            return "Single-Channel";
        case MAD_MODE_DUAL_CHANNEL:
            return "Dual-Channel";
        case MAD_MODE_JOINT_STEREO:
            return "Joint-Stereo";
        case MAD_MODE_STEREO:
            return "Stereo";
        default:
            return "?";
    }
}

enum mad_flow read_header(void *data, struct mad_header const * header)
{
    static long current_frame=0;
    static char current_file[PATH_MAX];

    char long_currenttime_str[13]; /* this *will* fill if you're using 100000+ minute mp3s */
    char long_remaintime_str[13];
    
    int new_file = 0;

    buffer *playbuf = (buffer *)data;
    mad_timer_t time_remaining;
    
    if (stop_playing_file)
    {
        stop_playing_file = 0;
        return MAD_FLOW_STOP;
    }
    
    if(strncmp(current_file, playbuf->filename, PATH_MAX) != 0)
    {
        new_file = 1;
        current_frame = 0;
        strncpy(current_file, playbuf->filename, PATH_MAX);
        current_file[PATH_MAX-1] = '\0';
    }

    current_frame++;

    if(options.opt & MPG321_VERBOSE_PLAY)
    {
        mad_timer_string(current_time, long_currenttime_str, "%.2u:%.2u.%.2u", MAD_UNITS_MINUTES,
                            MAD_UNITS_CENTISECONDS, 0);
        mad_timer_negate(&current_time);
        mad_timer_set(&time_remaining, playbuf->duration.seconds, playbuf->duration.fraction, 0);
        mad_timer_add(&time_remaining, current_time);
        mad_timer_negate(&current_time);

        mad_timer_string(time_remaining, long_remaintime_str, "%.2u:%.2u.%.2u", MAD_UNITS_MINUTES,
                            MAD_UNITS_CENTISECONDS, 0);
    }
                        
    /* Add the mad_timer after printing out the current information in verbose mode.
       The theory is, you're currently at the end of a frame, so you display the timestamp
       at the end of that frame (ie, before_time + length_of_last_frame) - then, increment
       the current time by the duration of the current frame, and season to taste */
    mad_timer_add(&current_time, header->duration);
    
    if (options.seek && current_frame < options.seek)
        return MAD_FLOW_IGNORE;
    
    if (new_file)
    {
        if (options.opt & MPG321_VERBOSE_PLAY)/*zip it good*/
        {
            fprintf(stderr, "MPEG 1.0, Layer: %s, Freq: %d, mode: %s, modext: %d, BPF : %d\n"
                    "Channels: %d, copyright: %s, original: %s, CRC: %s, emphasis: %d.\n"
                    "Bitrate: %d Kbits/s, Extension value: %d\n"
                    "Audio: 1:1 conversion, rate: %d, encoding: signed 16 bit, channels: %d\n",
                    layerstring(header->layer), header->samplerate, modestringucase(header->mode), header->mode_extension, 
                            header->bitrate*mad_timer_count(header->duration, MAD_UNITS_SECONDS)/8,
                    MAD_NCHANNELS(header), header->flags & MAD_FLAG_COPYRIGHT ? "Yes" : "No",
                    header->flags & MAD_FLAG_ORIGINAL ? "Yes" : "No", header->flags & MAD_FLAG_PROTECTION ? "Yes" : "No",
                    header->emphasis, header->bitrate/1000, header->mode_extension,
                    header->samplerate, MAD_NCHANNELS(header));
            }
        else if (!(options.opt & MPG321_QUIET_PLAY))/*I love Joey*/
        {
            fprintf(stderr, "MPEG 1.0 layer %s, %d kbit/s, %d Hz %s\n",
                layerstring(header->layer), header->bitrate/1000, header->samplerate, modestring(header->mode));
        }
    }
    
    if (options.opt & MPG321_VERBOSE_PLAY)
    {
        fprintf(stderr, "Frame# %5u [%5u], Time: %s [%s], \r", current_frame, 
                playbuf->num_frames - current_frame, long_currenttime_str, long_remaintime_str);
    }
    
    return MAD_FLOW_CONTINUE;
}        

/* Following two functions are adapted from mad_timer, from the 
   libmad distribution */
void scan(void const *ptr, ssize_t len,
		mad_timer_t *duration, unsigned long *numframes)
{
    struct mad_stream stream;
    struct mad_header header;

    mad_stream_init(&stream);
    mad_header_init(&header);

    mad_stream_buffer(&stream, ptr, len);

    *numframes = 0;

    while (1)
    {
        if (mad_header_decode(&header, &stream) == -1)
        {
            if (MAD_RECOVERABLE(stream.error))
                continue;
            else
                break;
        }

        (*numframes)++;

        mad_timer_add(duration, header.duration);
    }

    mad_header_finish(&header);
    mad_stream_finish(&stream);
}

int calc_length(char const *file, mad_timer_t *duration,
	 unsigned long *num_frames)
{
    int fd;
    struct stat stat;
    void *fdm;

    fd = open(file, O_RDONLY);
    if (fd == -1)
    {
        fprintf(stderr, "Unable to open file %s!\n",file);
        return -1;
    }

    if (fstat(fd, &stat) == -1)
    {
        fprintf(stderr, "Couldn't stat file %s!\n",file);
        close(fd);
        return -1;
    }

    if (!S_ISREG(stat.st_mode))
    {
        fprintf(stderr, "%s: Not a regular file\n", file);
        close(fd);
        return -1;
    }

    fdm = mmap(0, stat.st_size, PROT_READ, MAP_SHARED, fd, 0);
    if (fdm == MAP_FAILED)
    {
        fprintf(stderr, "mmap failed on %s!\n",file);
        close(fd);
        return -1;
    }

    scan(fdm, stat.st_size, duration, num_frames);

    if (munmap(fdm, stat.st_size) == -1)
    {
        fprintf(stderr, "munmap failed on %s!\n",file);
        close(fd);
        return -1;
    }

    if (close(fd) == -1)
    {
        fprintf(stderr, "couldn't close fd %d\n",fd);
        return -1;
    }

    return 0;
}

/* Scale and round samples to 16 bits. Taken directly from minimad.c,
   part of the libmad distribution. */
static inline 
signed int scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE; 

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}

/* Borrowed from SDL. There, the maximum volume is 128, which makes sense given that the
   sample is 8 bits. */
   
#define ADJUST_VOLUME(s,v)  ((s)*(v)/128)

enum mad_flow output(void *data,
                     struct mad_header const *header,
                     struct mad_pcm *pcm)
{
    register int nsamples = pcm->length;
    mad_fixed_t const *left_ch = pcm->samples[0], *right_ch = pcm->samples[1];
    
    static unsigned char stream[1152*4]; /* 1152 because that's what mad has as a max; *4 because
                                    there are 4 distinct bytes per sample (in 2 channel case) 
                                    */
    register char * ptr = stream;
    register signed int sample;

    /* We need to know information about the file before we can open the playdevice
       in some cases. So, we do it here. */
    if (!playdevice)
    {
        open_ao_playdevice(header);        
    }
    
    if (pcm->channels == 2)
    {
        while (nsamples--)
        {
            sample = ADJUST_VOLUME(scale(*left_ch++), options.volume);

#ifndef WORDS_BIGENDIAN
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
#else
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
#endif
            
            sample = ADJUST_VOLUME(scale(*right_ch++), options.volume);
#ifndef WORDS_BIGENDIAN
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
#else
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
#endif
        }

        ao_play(playdevice, stream, pcm->length * 4);
    }
    
    else
    {
        while (nsamples--)
        {
            sample = scale(*left_ch++);
            
#ifndef WORDS_BIGENDIAN
            *ptr++ = (unsigned char) (sample >> 0);
            *ptr++ = (unsigned char) (sample >> 8);
#else
            *ptr++ = (unsigned char) (sample >> 8);
            *ptr++ = (unsigned char) (sample >> 0);
#endif
        }
        ao_play(playdevice, stream, pcm->length * 2);
    }

    return MAD_FLOW_CONTINUE;        
}

#if 0
char *format_time_short(double t)
{
    int minutes, seconds;
    char * time;
    
    minutes = (int)(t / 60);
    seconds = (int)t-(60 * minutes);

    asprintf(&time,"%.2d:%.2d",minutes, (seconds>0)? seconds: 0);
    return time;
}
#endif
