/*
    mpg321 - a fully free clone of mpg123.
    Copyright (C) 2001 Joe Drew
    
    Originally based heavily upon:
    plaympeg - Sample MPEG player using the SMPEG library
    Copyright (C) 1999 Loki Entertainment Software
    
    Also uses some code from
    mad - MPEG audio decoder
    Copyright (C) 2000-2001 Robert Leslie
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _MPG321_H_
#define _MPG321_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ao/ao.h>
#include <mad.h>

/* playlist structure */
typedef struct pl
{
    char **files;
    int numfiles;
    int files_size;
    int random_play;
} playlist;

/* Private buffer for passing around with libmad */
typedef struct
{
    void * buf;
    int fd;
    ssize_t length;
    int done;
    unsigned long num_frames;
    mad_timer_t duration;
    char *filename;
} buffer;

typedef struct
{
    int opt;
    char *device;
    unsigned long seek;
    int volume;
} mpg321_options;    

extern mpg321_options options;
extern ao_device_t *playdevice;
extern mad_timer_t current_time;
extern int stop_playing_file;

enum
{
    MPG321_VERBOSE_PLAY  = 0x0001,
    MPG321_QUIET_PLAY    = 0x0002,

    MPG321_USE_STDOUT    = 0x0010,
    MPG321_USE_ALSA      = 0x0020,
    MPG321_USE_ESD       = 0x0040,
    MPG321_USE_NULL      = 0x0080,
    MPG321_USE_WAV       = 0x0100
};

#define DEFAULT_PLAYLIST_SIZE 1024
#define BUF_SIZE 4096 /* Size for read buffer for audio data */

/* playlist functions */
playlist * new_playlist();
void resize_playlist(playlist *pl);
char * get_next_file(playlist *pl);
void add_cmdline_files(playlist *pl, char *argv[]);
void load_playlist(playlist *pl, char *filename);
void set_random_play(playlist *pl);
void shuffle_files(playlist *pl);

/* network functions */
int tcp_open(char * address, int port);
int udp_open(char * address, int port);
int raw_open(char * arg);
int http_open(char * arg);
int ftp_open(char * arg);

/* libmad interfacing functions */
enum mad_flow read_from_mmap(void *data, struct mad_stream *stream);
enum mad_flow read_from_fd(void *data, struct mad_stream *stream);
enum mad_flow read_header(void *data, struct mad_header const * header);
enum mad_flow output(void *data, struct mad_header const *header, struct mad_pcm *pcm);
int calc_length(char const *file, mad_timer_t *duration, unsigned long *num_frames);

#endif /* _MPG321_H_ */
